package com.ibm.oti.reflect;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2010  All Rights Reserved
 */

import java.io.Serializable;
import java.lang.annotation.Annotation;
import java.lang.annotation.AnnotationFormatError;
import java.lang.annotation.AnnotationTypeMismatchException;
import java.lang.annotation.IncompleteAnnotationException;
import java.lang.reflect.Array;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.HashMap;

public class AnnotationParser {
	public static Annotation[] parseAnnotations(java.lang.reflect.Field field) {
		return parseAnnotations(getAnnotationsData(field), field.getDeclaringClass());
	}

	public static Annotation[] parseAnnotations(java.lang.reflect.Constructor constructor) {
		return parseAnnotations(getAnnotationsData(constructor), constructor.getDeclaringClass());
	}

	public static Annotation[] parseAnnotations(java.lang.reflect.Method method) {
		return parseAnnotations(getAnnotationsData(method), method.getDeclaringClass());
	}

	public static Annotation[][] parseParameterAnnotations(java.lang.reflect.Constructor constructor) {
		return parseParameterAnnotations(getParameterAnnotationsData(constructor), constructor.getDeclaringClass(), constructor.getParameterTypes().length);
	}

	public static Annotation[][] parseParameterAnnotations(java.lang.reflect.Method method) {
		return parseParameterAnnotations(getParameterAnnotationsData(method), method.getDeclaringClass(), method.getParameterTypes().length);
	}

	public static Object parseDefaultValue(java.lang.reflect.Method method) {
		byte[] elementValueData = getDefaultValueData(method);
		if (elementValueData == null) return null;
		ByteBuffer buf = ByteBuffer.wrap(elementValueData);
		return elementValueAsReturnType(method, parseElementValue(buf, method.getDeclaringClass()));
	}

	public static Annotation[] parseAnnotations(byte[] annotationsData, Class<?> declaringClass) {
		if (annotationsData == null) return new Annotation[0];
		ByteBuffer buf = ByteBuffer.wrap(annotationsData);
		int annotationsCount = buf.getChar();
		int j = 0;
		Annotation[] annotations = new Annotation[annotationsCount];
		for (int i = 0; i < annotationsCount; i++) {
			try {
				Annotation a = parseAnnotation(buf, declaringClass);
				annotations[j++] = a;
			} catch (TypeNotPresentException e) {
				/* Missing annotation classes should be ignored */
			}
		}
		if (j < annotationsCount) {
			Annotation tmp[] = new Annotation[j];
			System.arraycopy(annotations, 0, tmp, 0, j);
			annotations = tmp;
		}
		return annotations;
	}

	private static native byte[] getAnnotationsData(java.lang.reflect.Field field);
	private static native byte[] getAnnotationsData(java.lang.reflect.Constructor constructor);
	private static native byte[] getAnnotationsData(java.lang.reflect.Method method);
	private static native byte[] getParameterAnnotationsData(java.lang.reflect.Constructor constructor);
	private static native byte[] getParameterAnnotationsData(java.lang.reflect.Method method);
	private static native byte[] getDefaultValueData(java.lang.reflect.Method method);

	private static native int getIntAt(Object clazz, int cpIndex);
	private static native long getLongAt(Object clazz, int cpIndex);
	private static native float getFloatAt(Object clazz, int cpIndex);
	private static native double getDoubleAt(Object clazz, int cpIndex);
	private static native String getStringAt(Object clazz, int cpIndex);
	private static native String getUTF8At(Object clazz, int cpIndex);

	private static Annotation[][] parseParameterAnnotations(byte[] annotationsData, Class<?> declaringClass, int parametersCount) {
		if (annotationsData == null) {
			Annotation[][] annotations = new Annotation[parametersCount][];
			for (int i = 0; i < parametersCount; i++) {
				annotations[i] = new Annotation[0];
			}
			return annotations;
		}
		ByteBuffer buf = ByteBuffer.wrap(annotationsData);
		// TODO check whether the array length can be different from the actual parametersCount
		parametersCount = ((int) buf.get()) & 0xff;
		Annotation[][] annotations = new Annotation[parametersCount][];
		for (int i = 0; i < parametersCount; i++) {
			int annotationsCount = buf.getChar();
			int k = 0;
			annotations[i] = new Annotation[annotationsCount];
			for (int j = 0; j < annotationsCount; j++) {
				try {
					Annotation a = parseAnnotation(buf, declaringClass);
					annotations[i][k++] = a;
				} catch (TypeNotPresentException e) {
					/* Missing annotation classes should be ignored */
				}
			}
			if (k < annotationsCount) {
				Annotation tmp[] = new Annotation[k];
				System.arraycopy(annotations[i], 0, tmp, 0, k);
				annotations[i] = tmp;
			}
		}
		return annotations;
	}

	@SuppressWarnings("unchecked")
	private static Annotation parseAnnotation(ByteBuffer buf, Class<?> declaringClass) {
		Annotation result;
		int typeIndex = buf.getChar();
		String className = getUTF8At(declaringClass, typeIndex);
		className = className.substring(1, className.length() - 1).replace('/', '.');
		try {
			Class<? extends Annotation> forProxy = (Class<? extends Annotation>) Class.forName(className, true, declaringClass.getClassLoader());
			int elementValuePairCount = buf.getChar();
			HashMap<String, Object> elementValuePairs = new HashMap<String, Object>(); // TODO perf opportunity? elementValuePairCount * ?
			for (int i = 0; i < elementValuePairCount; i++) {
				int elementNameIndex = buf.getChar();
				String elementName = getUTF8At(declaringClass, elementNameIndex);
				Object elementValue = parseElementValue(buf, declaringClass);
				elementValuePairs.put(elementName, elementValue);
			}
			result = (Annotation) Proxy.newProxyInstance(forProxy.getClassLoader(),
					new Class[] { forProxy },
					new AnnotationInvocationHandler(forProxy, elementValuePairs));
		} catch (ClassNotFoundException e) {
			throw new TypeNotPresentException(className, e);
		}
		return result;
	}

	private static Object parseElementValue(ByteBuffer buf, Class<?> declaringClass) {
		byte tag = buf.get();
		switch (tag) {
		case 'B': return (byte) getIntAt(declaringClass, buf.getChar());
		case 'C': return (char) getIntAt(declaringClass, buf.getChar());
		case 'D': return getDoubleAt(declaringClass, buf.getChar());
		case 'F': return getFloatAt(declaringClass, buf.getChar());
		case 'I': return getIntAt(declaringClass, buf.getChar());
		case 'J': return getLongAt(declaringClass, buf.getChar());
		case 'S': return (short) getIntAt(declaringClass, buf.getChar());
		case 'Z': return getIntAt(declaringClass, buf.getChar()) != 0;
		case 's': return getStringAt(declaringClass, buf.getChar());
		case 'e': return parseEnum(buf, declaringClass);
		case 'c': return parseClass(buf, declaringClass);
		case '[': return parseArray(buf, declaringClass);
		case '@':
			try {
				return parseAnnotation(buf, declaringClass);
			} catch (TypeNotPresentException e) {
				return new TypeNotPresentElement(e.typeName(), e.getCause());
			}
		default: throw new AnnotationFormatError(""); // TODO error message
		}
	}

	private static Object parseArray(ByteBuffer buf, Class<?> declaringClass) {
		int valueCount = buf.getChar();
		Object[] values = new Object[valueCount];
		for (int i = 0; i < values.length; i++) {
			values[i] = parseElementValue(buf, declaringClass);
		}
		return values;
	}

	private static Object parseClass(ByteBuffer buf, Class<?> declaringClass) {
		int cpIndex = buf.getChar();
		String returnDescriptor = getUTF8At(declaringClass, cpIndex);

		if (returnDescriptor.length() == 1) {
			switch (returnDescriptor.charAt(0)) {
			case 'B': return Byte.TYPE;
			case 'C': return Character.TYPE;
			case 'D': return Double.TYPE;
			case 'F': return Float.TYPE;
			case 'I': return Integer.TYPE;
			case 'J': return Long.TYPE;
			case 'S': return Short.TYPE;
			case 'Z': return Boolean.TYPE;
			case 'V': return Void.TYPE;
			default: throw new AnnotationFormatError(""); // TODO error message
			}
		}
		String className = returnDescriptor.replace('/', '.');
		if (returnDescriptor.charAt(0) != '[') {
			if (className.charAt(0) != 'L' || className.charAt(className.length() - 1) != ';') {
				throw new AnnotationFormatError(""); // TODO error message
			}
			className = className.substring(1, className.length() - 1);
		}
		try {
			return Class.forName(className, true, declaringClass.getClassLoader());
		} catch (ClassNotFoundException e) {
			return new TypeNotPresentElement(className, e);
		}
	}

	@SuppressWarnings("unchecked")
	private static Object parseEnum(ByteBuffer buf, Class<?> declaringClass) {
		int typeNameIndex = buf.getChar();
		int constNameIndex = buf.getChar();
		String typeName = getUTF8At(declaringClass, typeNameIndex);
		String constName = getUTF8At(declaringClass, constNameIndex);

		// Turn the internal binary name of the type into a Java binary name.
		if (typeName.charAt(0) != 'L' || typeName.charAt(typeName.length() - 1) != ';') {
			throw new AnnotationFormatError(""); // TODO error message
		}

		String javaTypeName = typeName.substring(1, typeName.length() - 1).replace('/', '.');

		Class<? extends Enum> enumType = null;
		try {
			enumType = (Class<? extends Enum>) Class.forName(javaTypeName, true, declaringClass.getClassLoader());
			return Enum.valueOf(enumType, constName);
		} catch (ClassNotFoundException e) {
			return new TypeNotPresentElement(javaTypeName, e);
		} catch (IllegalArgumentException e) {
			return new EnumConstantNotPresentElement(enumType, constName);
		}
	}

	private static abstract class InvalidElement {
		abstract void throwException();
	}

	private static class TypeNotPresentElement extends InvalidElement {
		private String typeName;
		private Throwable throwable;

		TypeNotPresentElement(String typeName, Throwable throwable) {
			this.typeName = typeName;
			this.throwable = throwable;
		}

		@Override
		void throwException() {
			throw new TypeNotPresentException(typeName, throwable);
		}
	}

	private static class EnumConstantNotPresentElement extends InvalidElement {
		private Class<? extends Enum> enumType;
		private String constName;

		EnumConstantNotPresentElement(Class<? extends Enum> enumType, String constName) {
			this.enumType = enumType;
			this.constName = constName;
		}

		@Override
		void throwException() {
			throw new EnumConstantNotPresentException(enumType, constName);
		}
	}

	private static class AnnotationInvocationHandler implements InvocationHandler, Serializable {
		private static final long serialVersionUID = -5623887444054986634L;

		private HashMap<String, Object> elements;
		private Class<? extends Annotation> annType;

	AnnotationInvocationHandler(Class<? extends Annotation> annType, HashMap<String, Object> elements) {
		this.elements = elements;
		this.annType = annType;
	}

	public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
		final String methodName = method.getName();
		if (methodName.equals("toString")) { //$NON-NLS-1$
			return objectToString();
		} else if (methodName.equals("annotationType")) { //$NON-NLS-1$
			return annType;
		} else if (methodName.equals("equals")) { //$NON-NLS-1$
			return compare(args) ? Boolean.TRUE : Boolean.FALSE;
		} else if (methodName.equals("hashCode")) { //$NON-NLS-1$
			return new Integer(getHashValue());
		}

		Object elementValue = elements.get(methodName);
		if (elementValue == null) {
			elementValue = method.getDefaultValue();
			if (elementValue == null) {
				throw new IncompleteAnnotationException(annType, methodName);
			}
			return elementValue;
		}
		Object result = elementValueAsReturnType(method, elementValue);
		if (result != elementValue) elements.put(methodName, result);
		if (result.getClass().isArray()) {
			Object copy = Array.newInstance(result.getClass().getComponentType(), Array.getLength(result));
			System.arraycopy(result, 0, copy, 0, Array.getLength(result));
			result = copy;
		}
		return result;
	}

	private String objectToString()  throws Throwable {
		Method[] methods = annType.getDeclaredMethods();
		StringBuilder output = new StringBuilder(methods.length * 48);
		output.append('@').append(annType.getCanonicalName()).append('(');
		for (int i = 0; i < methods.length; i++) {
			output.append(methods[i].getName()).append('=');
			try {
				Object result = invoke(this, methods[i], new Object[0]);
				if (result.getClass().isArray()) {
					output.append(getArrayValue(result));
				} else {
					output.append(result);
				}
			} catch (TypeNotPresentException e) {
				Object result = elements.get(methods[i].getName());
				if (result == null) {
					result = methods[i].getDefaultValue();
				}
				output.append(result);
			}
			if (i!=methods.length-1) {
				output.append(", "); //$NON-NLS-1$
			}
		}
		output.append(')');

		return output.toString();
	}

	private boolean compare(Object[] args) throws Throwable {
		if (args.length != 1) {
			return false;
		}
		if (!(args[0] instanceof Annotation)) {
			return false;
		}

		Class annType2 = ((Annotation)args[0]).annotationType();
		if (!annType.equals(annType2)) {
			return false;
		}

		Method[] m = annType.getDeclaredMethods();
		for (int i = 0; i < m.length; i++) {
			Object result1 = invoke(this, m[i], new Object[0]);
			m[i].setAccessible(true);
			Object result2 = m[i].invoke(args[0], new Object[0]);

			if (result1 == null || result2 == null) {
				// handle null, although it shouldn't happen
				if (result1 != result2) return false;
			} else if (result1.getClass().isArray() && result2.getClass().isArray()) {
				if (result1 instanceof boolean[] && result2 instanceof boolean[]) {
					if (!Arrays.equals((boolean[])result1, (boolean[])result2))
						return false;
				} else if (result1 instanceof char[] && result2 instanceof char[]) {
					if (!Arrays.equals((char[])result1, (char[])result2))
						return false;
				} else if (result1 instanceof byte[] && result2 instanceof byte[]) {
					if (!Arrays.equals((byte[])result1, (byte[])result2))
						return false;
				} else if (result1 instanceof short[] && result2 instanceof short[]) {
					if (!Arrays.equals((short[])result1, (short[])result2))
						return false;
				} else if (result1 instanceof int[] && result2 instanceof int[]) {
					if (!Arrays.equals((int[])result1, (int[])result2))
						return false;
				} else if (result1 instanceof float[] && result2 instanceof float[]) {
					if (!Arrays.equals((float[])result1, (float[])result2))
						return false;
				} else if (result1 instanceof double[] && result2 instanceof double[]) {
					if (!Arrays.equals((double[])result1, (double[])result2))
						return false;
				} else if (result1 instanceof long[] && result2 instanceof long[]) {
					if (!Arrays.equals((long[])result1, (long[])result2))
						return false;
				} else if (result1 instanceof Object[] && result2 instanceof Object[]) {
					if (!Arrays.equals((Object[])result1, (Object[])result2))
						return false;
				} else {
					return false;
				}
			} else {
				if (!result1.equals(result2)) {
					return false;
				}
			}
		}
		return true;
	}

	private int getHashValue() throws Throwable {
		Method[] m = annType.getDeclaredMethods();
		int hashCode = 0;
		for (int i = 0; i < m.length; i++) {
			int tempHashCode = m[i].getName().hashCode() * 127;
			Object result = invoke(this, m[i], new Object[0]);
			if (result == null) {
				// handle null, although it shouldn't happen
				tempHashCode = 0;
			} else if (result.getClass().isArray()) {
				if (result instanceof boolean[]) {
					tempHashCode ^= Arrays.hashCode((boolean[])result);
				} else if (result instanceof char[]) {
					tempHashCode ^= Arrays.hashCode((char[])result);
				} else if (result instanceof byte[]) {
					tempHashCode ^= Arrays.hashCode((byte[])result);
				} else if (result instanceof short[]) {
					tempHashCode ^= Arrays.hashCode((short[])result);
				} else if (result instanceof int[]) {
					tempHashCode ^= Arrays.hashCode((int[])result);
				} else if (result instanceof float[]) {
					tempHashCode ^= Arrays.hashCode((float[])result);
				} else if (result instanceof double[]) {
					tempHashCode ^= Arrays.hashCode((double[])result);
				} else if (result instanceof long[]) {
					tempHashCode ^= Arrays.hashCode((long[])result);
				} else if (result instanceof Object[]) {
					tempHashCode ^= Arrays.hashCode((Object[])result);
				}
			} else {
				tempHashCode ^= result.hashCode();
			}
			hashCode += tempHashCode;
		}
		return hashCode;
	}

	private String getArrayValue(Object obj) {
		if (obj instanceof boolean[]) {
			return getBooleanArrayData((boolean[])obj);
		}else if (obj instanceof char[]) {
			return getCharArrayData((char[])obj);
		}else if (obj instanceof byte[]) {
			return getByteArrayData((byte[])obj);
		}else if (obj instanceof char[]) {
			return getCharArrayData((char[])obj);
		}else if (obj instanceof short[]) {
			return getShortArrayData((short[])obj);
		}else if (obj instanceof int[]) {
			return getIntArrayData((int[])obj);
		}else if (obj instanceof float[]) {
			return getFloatArrayData((float[])obj);
		}else if (obj instanceof double[]) {
			return getDoubleArrayData((double[])obj);
		}else if (obj instanceof long[]) {
			return getLongArrayData((long[])obj);
		}
		return getObjectArrayData((Object[])obj);
	}
	private String getObjectArrayData(Object[] objects) {
		StringBuilder output = new StringBuilder(objects.length * 20);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	private String getBooleanArrayData(boolean[] objects) {
		StringBuilder output = new StringBuilder(2 + objects.length * 6);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	private String getCharArrayData(char[] objects) {
		StringBuilder output = new StringBuilder(2 + objects.length * 4);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	private String getByteArrayData(byte[] objects) {
		StringBuilder output = new StringBuilder(2 + objects.length * 4);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	private String getShortArrayData(short[] objects) {
		StringBuilder output = new StringBuilder(objects.length * 6);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	private String getIntArrayData(int[] objects) {
		StringBuilder output = new StringBuilder(objects.length * 12);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	private String getFloatArrayData(float[] objects) {
		StringBuilder output = new StringBuilder(objects.length * 15);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	private String getDoubleArrayData(double[] objects) {
		StringBuilder output = new StringBuilder(objects.length * 20);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	private String getLongArrayData(long[] objects) {
		StringBuilder output = new StringBuilder(objects.length * 20);
		output.append('[');
		for (int i = 0; i < objects.length; i++) {
			output.append(objects[i]);
			if (i < objects.length - 1) {
				output.append(',');
			}
		}
		output.append(']');
		return output.toString();
	}
	}

	static Object elementValueAsReturnType(java.lang.reflect.Method method, Object elementValue) {
		Class<?> returnType = method.getReturnType();
		if (elementValue instanceof InvalidElement) ((InvalidElement)elementValue).throwException();
		if (returnType.isArray() && elementValue.getClass() == Object[].class) {
			Object[] array = (Object[]) elementValue;
			for (int i = 0; i < array.length; i++) {
				Object	elem = array[i];
				if (elem instanceof InvalidElement) ((InvalidElement)elem).throwException();
			}
			if (! returnType.isAssignableFrom(elementValue.getClass())) {
				Object result = Array.newInstance(returnType.getComponentType(), array.length);
				for (int i = 0; i < array.length; i++) {
					Array.set(result, i, array[i]);
				}
				return result;
			}
		}
		if (! returnType.isAssignableFrom(elementValue.getClass())) {
			if (returnType.isPrimitive()) {
				Class<?> valueType = elementValue.getClass();
				if (returnType == Boolean.TYPE && valueType == Boolean.class ||
					returnType == Character.TYPE && valueType == Character.class ||
					returnType == Byte.TYPE && valueType == Byte.class ||
					returnType == Short.TYPE && valueType == Short.class ||
					returnType == Integer.TYPE && valueType == Integer.class ||
					returnType == Long.TYPE && valueType == Long.class ||
					returnType == Float.TYPE && valueType == Float.class ||
					returnType == Double.TYPE && valueType == Double.class) {
					return elementValue;
				}
			}
			throw new AnnotationTypeMismatchException(method, elementValue.getClass().getName());
		}
		return elementValue;
	}
}
